{-# LANGUAGE
  StrictData #-}

module Little.Earley.Internal.Result where

import Data.Foldable (minimumBy)
import Data.Ord (comparing)

import Little.Earley.Internal.Core
import Little.Earley.Internal.Tree
import Little.Earley.Internal.Pretty
import Little.Earley.Internal.Render

-- | Result of 'parse'.
data Result n t c
  = ParseError
  | ParseSuccess (ParsedResult n t c)

-- | Successful result of 'parse'.
data ParsedResult n t c = ParsedResult
  { prParseTree :: TreeSet n t c
    -- ^ The resulting parse tree.
  , prAmbiguities :: [LocAmbiguity n t c]
    -- ^ A stream of localized ambiguities, when the parse tree is ambiguous.
  }

instance (PrettyPrint n, PrettyPrint c) => PrettyPrint (Result n t c) where
  prettyPrint ParseError = "Parse error"
  prettyPrint (ParseSuccess r) = unlines (pTree ++ pAmbig) where
    pTree = prettyTree (arbTree (prParseTree r))
    pAmbig = case prAmbiguities r of
      [] -> []
      ambs ->
        let (rg, Ambiguity t1 t2) = minimumBy (comparing fst) (take 1000 ambs)
            pRange = "Ambiguous parse between tokens number "
              ++ show (rangePos rg) ++ " and " ++ show (rangePos rg + rangeLen rg) ++ ":" in
        "" : pRange : "" : prettyTree t1 ++ "" : prettyTree t2

-- | Parse a chain of tokens @[c]@ given a grammar and a starting symbol @n@.
--
-- == Example
--
-- @
-- 'parse' arithG SUM \"1+2*3\"
-- @
parse :: (Ord n, Ord t) => Grammar n t c -> n -> [c] -> Result n t c
parse g n cs =
  case parseTreeSet g n cs of
    Nothing -> ParseError
    Just t -> ParseSuccess (ParsedResult
      { prParseTree = t
      , prAmbiguities = ambiguities 0 (length cs) t
      })

-- | Wrapped 'parse' with a pretty-printed result. Use this in the REPL.
--
-- == Example
--
-- @
-- 'pparse' arithG SUM \"1+2*3\"
-- @
--
-- Output:
--
-- >      +-----+--SUM #1---+
-- >      |     |           |
-- >   SUM #0   |      +PRODUCT #1-+
-- >      |     |      |     |     |
-- > PRODUCT #0 | PRODUCT #0 |     |
-- >      |     |      |     |     |
-- >  FACTOR #0 |  FACTOR #0 | FACTOR #0
-- >      |     |      |     |     |
-- >  NUMBER #0 |  NUMBER #0 | NUMBER #0
-- >      |     |      |     |     |
-- > -----------------------------------
-- >      1     +      2     *     3
pparse :: (Ord n, Ord t) => Grammar n t c -> n -> [c] -> Pretty (Result n t c)
pparse g n cs = Pretty (parse g n cs)
